/* 
 * vpopq_client
 * reads a checkpassword-token (http://cr.yp.to/checkpwd/interface.html) and
 * sends the given e-mail address to a vpopq_server, then exits with specific errorcodes
 *
 * this is to be used by qmail with spamcontrol >= 2.5.23 (qmail-recipients) patch
 *  
 * exit-codes:
 *  0: address ok
 *  1: address does not exist
 *  2: usage error
 *  111: temporary error (e.g. connection to server timed out)
 *
 * <rs@bytecamp.net> R. Schulze, bytecamp GmbH
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>

#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>

static const char __version[]="vpopq_client/0.5/20090319 <rs@bytecamp.net>";

#define	EXIT_ERROR 	111
#define EXIT_FOUND	0
#define EXIT_NOT_FOUND	1
#define EXIT_USAGE	2

#define FD_READ_PWDINFO	3

#define CHKPWD_RBUF	512
#define NET_TIMEOUT	2

#define WAIT_IO_WRITE 10
#define WAIT_IO_READ 11

char *checkpassword();
int dialog(char *, int, const char *);
int wait_for_io(int, int, int);

/* ****************************************************************************
 *************************************************************************** */

int main(int argc, const char* argv[])
{
if(argc!=3)
	{
	errx(EXIT_USAGE,"usage: %s ip-addr port\nreads checkpassword-info from fd %d\nasks server at host:port if address is existent",argv[0],FD_READ_PWDINFO);
	}

char *address=checkpassword();
if(address==NULL)
	{
	fprintf(stderr,"vpopq_client: no address in checkpassword-token\n");
	return EXIT_ERROR;
	}

fprintf(stderr,"vpopq_client: query for address %s\n",address);

int answer=dialog((char *)argv[1],atoi(argv[2]),address);
free(address);

/* 
 * return NOT FOUND only on successful negative lookup
 * in all other cases, let the mail pass to the next server
 */

if(answer==0)
        {
        return EXIT_NOT_FOUND;
        }
return EXIT_FOUND;
}


/* ****************************************************************************
 * dialog
 * asks the server for the address
 * returns 0 on not found, 1 on found, -1 on error
 *************************************************************************** */


int dialog(char *host, int port, const char *address)
{
int sock=socket(PF_INET,SOCK_STREAM,0);
if(sock==-1)
	{
	warn("socket: ");
	return -1;
	}

/*
 * enable non-blocking socket
 */

int sock_flags=fcntl(sock,F_GETFL);
if(sock_flags==-1)
	{
	warn("unable to fcntl()");
	close(sock);
	return -1;
	}
if(fcntl(sock,F_SETFL,sock_flags|O_NONBLOCK)==-1)
	{
	warn("unable to set socket to non-blocking mode");
	close(sock);
	return -1;
	}

/*
 * connect to host 
 */

struct sockaddr_in addr;

addr.sin_family=AF_INET; 
addr.sin_port=htons(port); 
addr.sin_addr.s_addr=inet_addr(host); 

int res=connect(sock,(struct sockaddr *)&addr,sizeof(addr));
if( (res==-1) && (errno!=EINPROGRESS) )
	{
	warn("connect: ");
	close(sock);
	return -1;
	}

/*
 * wait until socket is ready for writing (or NET_TIMEOUT reached)
 * write address, shutdown writer
 */
int w_ok=wait_for_io(sock,NET_TIMEOUT,WAIT_IO_WRITE);
if(w_ok<1)
	{
	warn((w_ok==0) ? "write timeout" : NULL);
	close(sock);
	return -1;
	}
	
if(write(sock,address,strlen(address))!=strlen(address))
	{
	warn("incomplete write: ");
	close(sock);
	return -1;
	}
if(write(sock,"\n",1)!=1)
	{
	warn("incomplete write: ");
	close(sock);
	return -1;
	}

shutdown(sock,SHUT_WR);

/*
 * wait until socket is ready for reading (or NET_TIMEOUT reached)
 * read answer, close socket
 */
int r_ok=wait_for_io(sock,NET_TIMEOUT,WAIT_IO_READ);
if(r_ok<1)
	{
	warn((r_ok==0) ? "read timeout" : NULL);
	close(sock);
	return -1;
	}
char answer;
if(read(sock,&answer,1)<0)
	{
	warn("incomplete read: ");
	}
close(sock);

switch(answer)
	{
	case '0': return 0; break;
	case '1': return 1; break;
	default: return -1; break;
	}
}

/* ****************************************************************************
 * wait_for_io
 * waits until a specific FD is readable/writeable or timeout is reached
 * ************************************************************************* */

int wait_for_io(int socket, int timeout, int io_mode)
{
struct timeval tv; 
fd_set myset; 

tv.tv_sec=NET_TIMEOUT; 
tv.tv_usec=0; 
FD_ZERO(&myset); 
FD_SET(socket, &myset); 

int select_res=(io_mode==WAIT_IO_WRITE) ? select(socket+1,NULL,&myset,NULL,&tv) : select(socket+1,&myset,NULL,NULL,&tv);

if(select_res==1)
	{
	/* select returned OK, lets check the socket */
			
	int sock_opt_val;
	socklen_t sock_opt_len=sizeof(int);

	if(getsockopt(socket,SOL_SOCKET,SO_ERROR,(void *)&sock_opt_val,&sock_opt_len)<0)
		{
		return -1;
		}
	return 1;
	}
else if(select_res==0)
	{
	return 0;
	}
return -1;
}

/* ****************************************************************************
 * checkpassword
 * implements checkpassword-interface
 * returns the address to be queried
 *************************************************************************** */

char *checkpassword(void)
{
unsigned int bytes_read=0;
unsigned int rd=0;

char finished=0;
char *buf=calloc(CHKPWD_RBUF,1);
unsigned int i=0;

while(!finished && (rd=read(FD_READ_PWDINFO,(buf+bytes_read),CHKPWD_RBUF)) )
	{
	/* 
         * find first occurance of \0 - address found
         */
	for(i=bytes_read; (i<(rd+bytes_read)) && (!finished); i++)
		{
		if(*(buf+i)==0)
			{
			finished=1;
			buf=realloc(buf,i+1);
			*(buf+i)=0;
			}
		}
	bytes_read+=rd;
	
	if(!finished) buf=realloc(buf,bytes_read+CHKPWD_RBUF);
	}
/*close(FD_READ_PWDINFO);*/
return buf;
}


